@@ -139,6 +139,7 @@ static void gconf_client_queue_notify       (GConfClient *client,
                                              const char  *key);
 static void gconf_client_flush_notifies     (GConfClient *client);
 static void gconf_client_unqueue_notifies   (GConfClient *client);
+static void notify_one_entry (GConfClient *client, GConfEntry  *entry);
 
 static GConfEntry* get (GConfClient  *client,
                         const gchar  *key,
@@ -745,6 +746,22 @@ gconf_client_notify_remove  (GConfClient* client,
     }
 }
 
+void
+gconf_client_notify (GConfClient* client, const char* key)
+{
+  GConfEntry *entry;
+
+  g_return_if_fail (client != NULL);
+  g_return_if_fail (GCONF_IS_CLIENT(client));
+  g_return_if_fail (key != NULL);
+
+  entry = gconf_client_get_entry (client, key, NULL, TRUE, NULL);
+  if (entry != NULL)
+    {
+      notify_one_entry (client, entry);
+    }
+}
+
 void
 gconf_client_set_error_handling(GConfClient* client,
                                 GConfClientErrorHandlingMode mode)
@@ -989,6 +1006,28 @@ gconf_client_unset          (GConfClient* client,
     return TRUE;
 }
 
+gboolean
+gconf_client_recursive_unset (GConfClient *client,
+                              const char     *key,
+                              GConfUnsetFlags flags,
+                              GError        **err)
+{
+  GError* error = NULL;
+
+  trace ("Unsetting '%s'\n", key);
+  
+  PUSH_USE_ENGINE (client);
+  gconf_engine_recursive_unset(client->engine, key, flags, &error);
+  POP_USE_ENGINE (client);
+  
+  handle_error(client, error, err);
+
+  if (error != NULL)
+    return FALSE;
+  else
+    return TRUE;
+}
+
 static GSList*
 copy_entry_list (GSList *list)
 {
@@ -2320,8 +2359,8 @@ notify_listeners_callback(GConfListeners* listeners,
                           gpointer listener_data,
                           gpointer user_data)
 {
-  struct ClientAndEntry* cae = user_data;
   Listener* l = listener_data;
+  struct ClientAndEntry* cae = user_data;
   
   g_return_if_fail (cae != NULL);
   g_return_if_fail (cae->client != NULL);
