@@ -45,6 +45,9 @@ static int recursive_list = FALSE;
 static int dump_values = FALSE;
 static int set_schema_mode = FALSE;
 static char* value_type = NULL;
+static int get_type_mode = FALSE;
+static int get_list_size_mode = FALSE;
+static int get_list_element_mode = FALSE;
 static char* value_list_type = NULL;
 static char* value_car_type = NULL;
 static char* value_cdr_type = NULL;
@@ -59,6 +62,7 @@ static char* entry_file = NULL;
 static const char* config_source = NULL;
 static int use_local_source = FALSE;
 static int makefile_install_mode = FALSE;
+static int makefile_uninstall_mode = FALSE;
 static int break_key_mode = FALSE;
 static int break_dir_mode = FALSE;
 static int short_docs_mode = FALSE;
@@ -215,7 +219,34 @@ struct poptOption options[] = {
     0,
     N_("Specify the type of the value being set, or the type of the value a schema describes. Unique abbreviations OK."),
     N_("int|bool|float|string|list|pair")
-  },  
+  },
+  {
+    "get-type",
+    'T',
+    POPT_ARG_NONE,
+    &get_type_mode,
+    0,
+    N_("Print the data type of a key to standard output."),
+    NULL
+  },
+  {
+    "get-list-size",
+    '\0',
+    POPT_ARG_NONE,
+    &get_list_size_mode,
+    0,
+    N_("Get the number of elements in a list key."),
+    NULL
+  },
+  {
+    "get-list-element",
+    '\0',
+    POPT_ARG_NONE,
+    &get_list_element_mode,
+    0,
+    N_("Get a specific element from a list key, numerically indexed."),
+    NULL
+  },
   { 
     "list-type",
     '\0',
@@ -306,6 +337,15 @@ struct poptOption options[] = {
     N_("Properly installs schema files on the command line into the database. GCONF_CONFIG_SOURCE environment variable should be set to a non-default config source or set to the empty string to use the default."),
     NULL
   },
+  {
+    "makefile-uninstall-rule",
+    '\0',
+    POPT_ARG_NONE,
+    &makefile_uninstall_mode,
+    0,
+    N_("Properly uninstalls schema files on the command line from the database. GCONF_CONFIG_SOURCE environment variable should be set to a non-default config source or set to the empty string to use the default."),
+    NULL
+  },
   {
     "break-key",
     '\0',
@@ -401,6 +441,7 @@ struct poptOption options[] = {
 static int do_break_key(GConfEngine* conf, const gchar** args);
 static int do_break_directory(GConfEngine* conf, const gchar** args);
 static int do_makefile_install(GConfEngine* conf, const gchar** args);
+static int do_makefile_uninstall(GConfEngine* conf, const gchar** args);
 static int do_recursive_list(GConfEngine* conf, const gchar** args);
 static int do_dump_values(GConfEngine* conf, const gchar** args);
 static int do_all_pairs(GConfEngine* conf, const gchar** args);
@@ -408,11 +449,14 @@ static void list_pairs_in_dir(GConfEngine* conf, const gchar* dir, guint depth);
 static int get_schema_from_xml(xmlNodePtr node, gchar **schema_key, GHashTable** schemas_hash, GSList **applyto_list);
 static int get_first_value_from_xml(xmlNodePtr node, GConfValue** ret_value);
 static void print_value_in_xml(GConfValue* value, int indent);
-static void dump_entries_in_dir(GConfEngine* conf, const gchar* dir);
+static void dump_entries_in_dir(GConfEngine* conf, const gchar* dir, const gchar *base_dir);
 static gboolean do_dir_exists(GConfEngine* conf, const gchar* dir);
 static void do_spawn_daemon(GConfEngine* conf);
 static int do_get(GConfEngine* conf, const gchar** args);
 static int do_set(GConfEngine* conf, const gchar** args);
+static int do_get_type(GConfEngine* conf, const gchar** args);
+static int do_get_list_size(GConfEngine* conf, const gchar** args);
+static int do_get_list_element(GConfEngine* conf, const gchar** args);
 static int do_set_schema(GConfEngine* conf, const gchar** args);
 static int do_all_entries(GConfEngine* conf, const gchar** args);
 static int do_unset(GConfEngine* conf, const gchar** args);
@@ -473,15 +517,28 @@ main (int argc, char** argv)
     }
 
   if ((set_mode && get_mode) ||
-      (set_mode && unset_mode))
+      (set_mode && unset_mode) ||
+      (set_mode && get_type_mode) ||
+      (set_mode && get_list_size_mode) ||
+      (set_mode && get_list_element_mode)) 
     {
       fprintf(stderr, _("Can't set and get/unset simultaneously\n"));
       return 1;
     }
 
+  if ((get_type_mode && set_mode) ||
+      (get_type_mode && unset_mode))
+    {
+      fprintf(stderr, _("Can't get type and set/unset simultaneously\n"));
+      return 1;
+    }
+
   if ((all_entries_mode && get_mode) ||
       (all_entries_mode && set_mode) ||
-      (all_entries_mode && unset_mode))
+      (all_entries_mode && get_type_mode) ||
+      (all_entries_mode && unset_mode) ||
+      (all_entries_mode && get_list_size_mode) ||
+      (all_entries_mode && get_list_element_mode))
     {
       fprintf(stderr, _("Can't use --all-entries with --get or --set\n"));
       return 1;
@@ -489,7 +546,10 @@ main (int argc, char** argv)
 
   if ((all_subdirs_mode && get_mode) ||
       (all_subdirs_mode && set_mode) ||
-      (all_subdirs_mode && unset_mode))
+      (all_subdirs_mode && get_type_mode) ||
+      (all_subdirs_mode && unset_mode) ||
+      (all_subdirs_mode && get_list_size_mode) ||
+      (all_subdirs_mode && get_list_element_mode))
     {
       fprintf(stderr, _("Can't use --all-dirs with --get or --set\n"));
       return 1;
@@ -498,6 +558,9 @@ main (int argc, char** argv)
   if ((recursive_list && get_mode) ||
       (recursive_list && set_mode) ||
       (recursive_list && unset_mode) ||
+      (recursive_list && get_type_mode) ||
+      (recursive_list && get_list_size_mode) ||
+      (recursive_list && get_list_element_mode) ||
       (recursive_list && all_entries_mode) ||
       (recursive_list && all_subdirs_mode))
     {
@@ -508,6 +571,9 @@ main (int argc, char** argv)
   if ((set_schema_mode && get_mode) ||
       (set_schema_mode && set_mode) ||
       (set_schema_mode && unset_mode) ||
+      (set_schema_mode && get_type_mode) ||
+      (set_schema_mode && get_list_size_mode) ||
+      (set_schema_mode && get_list_element_mode) ||
       (set_schema_mode && all_entries_mode) ||
       (set_schema_mode && all_subdirs_mode))
     {
@@ -529,7 +595,9 @@ main (int argc, char** argv)
 
   if (ping_gconfd && (shutdown_gconfd || set_mode || get_mode || unset_mode ||
                       all_subdirs_mode || all_entries_mode || recursive_list || 
-                      spawn_gconfd || dir_exists || schema_file || makefile_install_mode ||
+                      get_type_mode || get_list_size_mode || get_list_element_mode ||
+                      spawn_gconfd || dir_exists || schema_file ||
+                      makefile_install_mode || makefile_uninstall_mode ||
                       break_key_mode || break_dir_mode || short_docs_mode ||
                          long_docs_mode || schema_name_mode))
     {
@@ -539,7 +607,9 @@ main (int argc, char** argv)
 
   if (dir_exists && (shutdown_gconfd || set_mode || get_mode || unset_mode ||
                      all_subdirs_mode || all_entries_mode || recursive_list || 
-                     spawn_gconfd || schema_file || makefile_install_mode ||
+                     get_type_mode || get_list_size_mode || get_list_element_mode ||
+                     spawn_gconfd || schema_file ||
+                     makefile_install_mode || makefile_uninstall_mode ||
                      break_key_mode || break_dir_mode || short_docs_mode ||
                          long_docs_mode || schema_name_mode))
     {
@@ -549,7 +619,9 @@ main (int argc, char** argv)
 
   if (schema_file && (shutdown_gconfd || set_mode || get_mode || unset_mode ||
                       all_subdirs_mode || all_entries_mode || recursive_list || 
-                      spawn_gconfd || dir_exists || makefile_install_mode ||
+                      get_type_mode || get_list_size_mode || get_list_element_mode ||
+                      spawn_gconfd || dir_exists ||
+                      makefile_install_mode || makefile_uninstall_mode ||
                       break_key_mode || break_dir_mode || short_docs_mode ||
                          long_docs_mode || schema_name_mode))
     {
@@ -560,6 +632,8 @@ main (int argc, char** argv)
 
   if (makefile_install_mode && (shutdown_gconfd || set_mode || get_mode || unset_mode ||
                                 all_subdirs_mode || all_entries_mode || recursive_list || 
+                                get_type_mode || get_list_size_mode || get_list_element_mode ||
+                                makefile_uninstall_mode ||
                                 spawn_gconfd || dir_exists || schema_file ||
                                 break_key_mode || break_dir_mode || short_docs_mode ||
                          long_docs_mode || schema_name_mode))
@@ -567,12 +641,25 @@ main (int argc, char** argv)
       fprintf(stderr, _("--makefile-install-rule must be used by itself.\n"));
       return 1;
     }
-
-
+  
+  if (makefile_uninstall_mode && (shutdown_gconfd || set_mode || get_mode ||
+                                  unset_mode || all_subdirs_mode ||
+                                  all_entries_mode || recursive_list || 
+                                  makefile_install_mode ||
+                                  spawn_gconfd || dir_exists || schema_file ||
+                                  break_key_mode || break_dir_mode || short_docs_mode ||
+                                  long_docs_mode || schema_name_mode))
+    {
+      fprintf(stderr, _("--makefile-uninstall-rule must be used by itself.\n"));
+      return 1;
+    }
+  
   if (break_key_mode && (shutdown_gconfd || set_mode || get_mode || unset_mode ||
                                 all_subdirs_mode || all_entries_mode || recursive_list || 
+                                get_type_mode || get_list_size_mode || get_list_element_mode ||
                                 spawn_gconfd || dir_exists || schema_file ||
-                                makefile_install_mode || break_dir_mode || short_docs_mode ||
+                                makefile_install_mode || makefile_uninstall_mode ||
+                                break_dir_mode || short_docs_mode ||
                          long_docs_mode || schema_name_mode))
     {
       fprintf(stderr, _("--break-key must be used by itself.\n"));
@@ -582,8 +669,10 @@ main (int argc, char** argv)
   
   if (break_dir_mode && (shutdown_gconfd || set_mode || get_mode || unset_mode ||
                                 all_subdirs_mode || all_entries_mode || recursive_list || 
+                                get_type_mode || get_list_size_mode || get_list_element_mode ||
                                 spawn_gconfd || dir_exists || schema_file ||
-                                break_key_mode || makefile_install_mode || short_docs_mode ||
+                                break_key_mode || makefile_install_mode ||
+                                makefile_uninstall_mode || short_docs_mode ||
                          long_docs_mode || schema_name_mode))
     {
       fprintf(stderr, _("--break-directory must be used by itself.\n"));
@@ -638,8 +727,15 @@ main (int argc, char** argv)
       g_print (_("GCONF_DISABLE_MAKEFILE_SCHEMA_INSTALL is set, not installing schemas\n"));
       makefile_install_mode = FALSE;
     }
-  
-  if (makefile_install_mode)
+
+  if (makefile_uninstall_mode &&
+      g_getenv ("GCONF_DISABLE_MAKEFILE_SCHEMA_UNINSTALL"))
+    {
+      g_print (_("GCONF_DISABLE_MAKEFILE_SCHEMA_UNINSTALL is set, not uninstalling schemas\n"));
+      makefile_uninstall_mode = FALSE;
+    }
+
+  if (makefile_install_mode || makefile_uninstall_mode)
     {
       g_assert (config_source == NULL);
 
@@ -743,6 +839,16 @@ main (int argc, char** argv)
       return retval;
     }
 
+  if (makefile_uninstall_mode)
+    {
+      const gchar** args = poptGetArgs(ctx);
+      gint retval = do_makefile_uninstall (conf, args);
+      
+      gconf_engine_unref (conf);
+
+      return retval;
+    }
+
   if (break_key_mode)
     {
       const gchar** args = poptGetArgs(ctx);
@@ -784,6 +890,36 @@ main (int argc, char** argv)
         }
     }
 
+  if (get_type_mode)
+    {
+      const gchar** args = poptGetArgs (ctx);
+      if (do_get_type (conf, args) == 1)
+        {
+          gconf_engine_unref (conf);
+          return 1;
+        }
+    }
+
+  if (get_list_size_mode)
+    {
+      const gchar** args = poptGetArgs (ctx);
+      if (do_get_list_size (conf, args) == 1)
+        {
+          gconf_engine_unref (conf);
+          return 1;
+        }
+    }
+
+  if (get_list_element_mode)
+    {
+      const gchar** args = poptGetArgs (ctx);
+      if (do_get_list_element (conf, args) == 1)
+        {
+          gconf_engine_unref (conf);
+          return 1;
+        }
+    }
+
   if (set_schema_mode)
     {
       const gchar** args = poptGetArgs(ctx);
@@ -986,7 +1122,7 @@ do_recursive_list(GConfEngine* conf, const gchar** args)
 }
 
 static void 
-recurse_subdir_dump(GConfEngine* conf, GSList* subdirs)
+recurse_subdir_dump(GConfEngine* conf, GSList* subdirs, const gchar* base_dir)
 {
   GSList* tmp;
 
@@ -996,9 +1132,11 @@ recurse_subdir_dump(GConfEngine* conf, GSList* subdirs)
     {
       gchar* s = tmp->data;
       
-      dump_entries_in_dir(conf, s);
+      dump_entries_in_dir(conf, s, base_dir);
 
-      recurse_subdir_dump(conf, gconf_engine_all_dirs(conf, s, NULL));
+      recurse_subdir_dump(conf,
+			  gconf_engine_all_dirs(conf, s, NULL),
+			  base_dir);
 
       g_free(s);
       
@@ -1027,9 +1165,9 @@ do_dump_values(GConfEngine* conf, const gchar** args)
 
       subdirs = gconf_engine_all_dirs(conf, *args, NULL);
 
-      dump_entries_in_dir(conf, *args);
+      dump_entries_in_dir(conf, *args, *args);
           
-      recurse_subdir_dump(conf, subdirs);
+      recurse_subdir_dump(conf, subdirs, *args);
 
       printf("  </entrylist>\n");
  
@@ -1037,6 +1175,7 @@ do_dump_values(GConfEngine* conf, const gchar** args)
     }
 
   printf("</gconfentryfile>\n");
+  return 0;
 }
 
 static void 
@@ -1248,7 +1387,7 @@ print_value_in_xml(GConfValue* value, int indent)
       g_free(tmp);
       break;
     case GCONF_VALUE_STRING:
-      tmp = gconf_value_to_string(value);
+      tmp = g_markup_escape_text(gconf_value_get_string(value), -1);
       printf("%s  <string>%s</string>\n", whitespace, tmp);
       g_free(tmp);
       break;
@@ -1301,7 +1440,7 @@ get_key_relative(const gchar* key, const gchar* dir)
 }
 
 static void 
-dump_entries_in_dir(GConfEngine* conf, const gchar* dir)
+dump_entries_in_dir(GConfEngine* conf, const gchar* dir, const gchar* base_dir)
 {
   GSList* entries;
   GSList* tmp;
@@ -1325,12 +1464,12 @@ dump_entries_in_dir(GConfEngine* conf, const gchar* dir)
       printf("    <entry>\n");
 
       printf("      <key>%s</key>\n",
-             get_key_relative(gconf_entry_get_key(entry), dir));
+             get_key_relative(gconf_entry_get_key(entry), base_dir));
 
       /* <schema_key> will only be relative if its under the base dir */
       if (gconf_entry_get_schema_name(entry))
         printf("      <schema_key>%s</schema_key>\n",
-               get_key_relative(gconf_entry_get_schema_name(entry), dir));
+               get_key_relative(gconf_entry_get_schema_name(entry), base_dir));
 
       if (entry->value)
         print_value_in_xml(entry->value, 6);
@@ -1636,6 +1775,160 @@ do_set(GConfEngine* conf, const gchar** args)
   return 0;
 }
 
+static int
+do_get_type(GConfEngine* conf, const gchar** args)
+{
+  GError* err = NULL;
+
+  if (args == NULL)
+    {
+      fprintf (stderr, _("Must specify a key or keys to get type\n"));
+      return 1;
+    }
+      
+  while (*args)
+    {
+      GConfValue* value;
+
+      err = NULL;
+
+      value = gconf_engine_get (conf, *args, &err);
+         
+      if (value != NULL)
+	{
+	  printf("%s\n", gconf_value_type_to_string (value->type));
+	}
+      else
+        {
+          if (err == NULL)
+            {
+              fprintf (stderr, _("No value set for `%s'\n"), *args);
+            }
+          else
+            {
+              fprintf (stderr, _("Failed to get value for `%s': %s\n"),
+                      *args, err->message);
+              g_error_free (err);
+              err = NULL;
+            }
+        }
+ 
+      ++args;
+    }
+  return 0;
+}
+
+static int
+do_get_list_size(GConfEngine* conf, const gchar** args)
+{
+  GError* err = NULL;
+  GConfValue *list = NULL;
+
+  if (args == NULL || *args == NULL)
+    {
+      fprintf (stderr, _("Must specify a key to lookup size of.\n"));
+      return 1;
+    }
+
+  list = gconf_engine_get (conf, *args, &err);
+
+  if (list == NULL)
+    {
+      if (err == NULL)
+	{
+	  fprintf (stderr, _("No value set for `%s'\n"), *args);
+	}
+      else
+	{
+	  fprintf (stderr, _("Failed to get value for `%s': %s\n"),
+		  *args, err->message);
+	  g_error_free (err);
+	  err = NULL;
+	}
+
+      return 1;
+    }
+
+  if (list->type != GCONF_VALUE_LIST)
+    {
+      fprintf (stderr, _("Key %s is not a list.\n"), *args);
+      return 1;
+    }
+
+  printf ("%u\n", g_slist_length (gconf_value_get_list (list)));
+
+  return 0;
+}
+
+static int
+do_get_list_element(GConfEngine* conf, const gchar** args)
+{
+  GError* err = NULL;
+  GConfValue *list = NULL, *element = NULL;
+  GSList *iter = NULL;
+  gchar* s = NULL;
+  int idx = 0;
+
+  if (args == NULL || *args == NULL)
+    {
+      fprintf (stderr, _("Must specify a key from which to get list element.\n"));
+      return 1;
+    }
+
+  list = gconf_engine_get (conf, *args, &err);
+
+  if (list == NULL)
+    {
+      if (err == NULL)
+	{
+	  fprintf (stderr, _("No value set for `%s'\n"), *args);
+	}
+      else
+	{
+	  fprintf (stderr, _("Failed to get value for `%s': %s\n"),
+		  *args, err->message);
+	  g_error_free (err);
+	  err = NULL;
+	}
+
+      return 1;
+    }
+
+  if (list->type != GCONF_VALUE_LIST)
+    {
+      fprintf (stderr, _("Key %s is not a list.\n"), *args);
+      return 1;
+    }
+
+  if (args[1] == NULL)
+    {
+      fprintf (stderr, _("Must specify list index.\n"));
+      return 1;
+    }
+
+  idx = atoi (args[1]);
+  if (idx < 0)
+    {
+      fprintf (stderr, _("List index must be non-negative.\n"));
+      return 1;
+    }
+
+  iter = gconf_value_get_list (list);
+  element = g_slist_nth_data (iter, idx);
+
+  if (element == NULL)
+    {
+      fprintf (stderr, _("List index is out of bounds.\n"));
+      return 1;
+    }
+
+  s = gconf_value_to_string (element);
+  printf ("%s\n", s);
+  g_free (s);
+
+  return 0;
+} 
+
 enum
 {
   SCHEMA_INFO_SHORT_DOCS,
@@ -2996,8 +3289,12 @@ process_key_list(GConfEngine* conf, const gchar* schema_name, GSList* keylist)
   GSList* tmp;
   GError* error = NULL;
 
-  tmp = keylist;
+  if (makefile_uninstall_mode)
+    {
+      schema_name = NULL;
+    }
 
+  tmp = keylist;
   while (tmp != NULL)
     {
       if (!gconf_engine_associate_schema(conf, tmp->data, schema_name,  &error))
@@ -3133,6 +3430,38 @@ hash_install_foreach(gpointer key, gpointer value, gpointer user_data)
   gconf_schema_free(schema);
 }
 
+static void
+hash_uninstall_foreach(gpointer key, gpointer value, gpointer user_data)
+{
+  struct {
+    GConfEngine* conf;
+    char* key;
+  } *info;
+  GConfSchema* schema;
+  GError* error = NULL;
+  
+  info = user_data;
+  schema = value;
+
+  if (!gconf_engine_unset (info->conf, info->key, &error))
+    {
+      g_assert(error != NULL);
+
+      fprintf(stderr, _("WARNING: failed to uninstall schema `%s' locale `%s': %s\n"),
+              info->key, gconf_schema_get_locale(schema), error->message);
+      g_error_free(error);
+      error = NULL;
+    }
+  else
+    {
+      g_assert(error == NULL);
+      printf(_("Uninstalled schema `%s' from locale `%s'\n"),
+             info->key, gconf_schema_get_locale(schema));
+    }
+
+  gconf_schema_free(schema);
+}
+
 static int
 process_schema(GConfEngine* conf, xmlNodePtr node)
 {
@@ -3157,7 +3486,10 @@ process_schema(GConfEngine* conf, xmlNodePtr node)
 
       hash_foreach_info.conf = conf;
       hash_foreach_info.key = schema_key;
-      g_hash_table_foreach(schemas_hash, hash_install_foreach, &hash_foreach_info);
+      if (makefile_uninstall_mode)
+        g_hash_table_foreach(schemas_hash, hash_uninstall_foreach, &hash_foreach_info);
+      else 
+        g_hash_table_foreach(schemas_hash, hash_install_foreach, &hash_foreach_info);
     }
   else
     {
@@ -3334,6 +3666,38 @@ do_makefile_install(GConfEngine* conf, const gchar** args)
   return 0;
 }
 
+static int
+do_makefile_uninstall(GConfEngine* conf, const gchar** args)
+{
+  GError* err = NULL;
+  
+  if (args == NULL)
+    {
+      fprintf(stderr, _("Must specify some schema files to uninstall\n"));
+      return 1;
+    }
+
+  while (*args)
+    {
+      if (do_load_file(conf, LOAD_SCHEMA_FILE, *args, NULL) != 0)
+        return 1;
+
+      ++args;
+    }
+
+  gconf_engine_suggest_sync(conf, &err);
+
+  if (err != NULL)
+    {
+      fprintf(stderr, _("Error syncing config data: %s"),
+              err->message);
+      g_error_free(err);
+      return 1;
+    }
+  
+  return 0;
+}
+
 typedef enum {
   BreakageSetBadValues,
   BreakageCleanup
